/* FFT for IBM */

#include <malloc.h>
#include <math.h>
#include <mpi.h>
#include "fft.h"

typedef struct {
    int i;
    complex X;
} sendmax;

static int cid;

double* W;
complex* y;
complex* x;
complex* X;
int* ep;
complex* ew;
int* index;
complex* ww;

static int windex;

complex get_w(int);

int main(int argc, char **argv)
{
    int i,j,k,l,m,n,b,p,q,e,r;
    int vid;
    int md;
    int idx[8];
    double t,dt,sum;
    double max1,max2;
    int i1,i2;
    complex w,z;
    complex *bufs;
    complex *bufd;
    complex r0,r1,r2,r3;
    complex v[9],u[8],s[5];
    double d;
    sendmax sm[2];
    complex amax[2];
    int aindex[2];
    MPI_Status status;

    MPI_Init(&argc,&argv);
    MPI_Comm_size(MPI_COMM_WORLD,&n);
    MPI_Comm_rank(MPI_COMM_WORLD,&cid);

    MPI_Barrier(MPI_COMM_WORLD);

    W = (double *)malloc((N/4+1) * sizeof(complex));
    y = (complex *)malloc(N_PE * sizeof(complex));
    x = (complex *)malloc(N_PE * sizeof(complex));
    X = (complex *)malloc(N_PE * sizeof(complex));
    ep = (int *)malloc(N_PE/5 * sizeof(int));
    ew = (complex *)malloc(N_PE/2 * sizeof(complex));
    index = (int *)malloc(N_PE * sizeof(int));
    ww = (complex *)malloc(12 * N_PE * sizeof(complex));

    /* W $@$N7W;;(J */
    dt = 2.0 * PI / (double)N;
    for (i = 0 ; i <= N/4 ; i++) {
	t = dt * (double)i;
	W[i] = sin(t);
    }
    idx[0]=0; idx[1]=4; idx[2]=2; idx[3]=6;
    idx[4]=1; idx[5]=5; idx[6]=3; idx[7]=7;

    /* ep $@$N7W;;(J */
    q = NumberOfCell / 2; n = 1; e = 0;
    for (k = 0 ; k < 5 ; k++) {
	e += ((cid / q) % 2) * n / 2;
	q /= 2; n *= 2;
    }
    ep[0] = e;
    if (cid >= NumberOfCell / 2) {
	ep[0] += NumberOfCell / 2;
    }
    m = 1;
    for (k = 0 ; k < 4 ; k++) {	/* $@CJ$N?t(J */
	for (i = m - 1 ; i >= 0 ; i--) {
	    ep[i*8] = ep[i];
	    ep[i*8+1] = ep[i*8] + n * 4;
	    ep[i*8+2] = ep[i*8] + n * 2;
	    ep[i*8+3] = ep[i*8] + n * 6;
	    ep[i*8+4] = ep[i*8] + n;
	    ep[i*8+5] = ep[i*8] + n * 5;
	    ep[i*8+6] = ep[i*8] + n * 3;
	    ep[i*8+7] = ep[i*8] + n * 7;
	}
	n *= 8; m *= 8;
    }
    for (k = 0 ; k < N5_PE ; k++) {	/* $@CJ$N?t(J */
	if (k < N5_PE - 1) {
	    for (i = m - 1 ; i >= 0 ; i--) {
		ep[i*5] = ep[i];
		ep[i*5+1] = ep[i*5] + n;
		ep[i*5+2] = ep[i*5] + n * 2;
		ep[i*5+3] = ep[i*5] + n * 3;
		ep[i*5+4] = ep[i*5] + n * 4;
	      }
	}
	m *= 5;	n *= 5;
    }

    /* index $@$N7W;;(J */
    for (i = 0 ; i < N_PE ; i++) {
	n = (cid % (NumberOfCell / 2)) * 2;
	if (cid >= NumberOfCell / 2) {
	    n++;
	}
	index[i] = i + n * N_PE;
	/* $@=gJQ99(J */
	n = 1; p = N;
	for (k = 0 ; k < N5 ; k++) {	/* $@CJ$N?t(J */
	    n *= 5; p /= 5;
	    j = index[i] % 5;
	    index[i] = (index[i] - j) / 5;
	    j += index[i] / p * 5;
	    index[i] = index[i] % p;
	    index[i] = j*p+index[i];
	}
	for (k = 0 ; k < N2 ; k++) {	/* $@CJ$N?t(J */
	    n *= 2; p /= 2;
	    j = index[i] % 2;
	    index[i] = (index[i] - j) / 2;
	    j += index[i] / p * 2;
	    index[i] = index[i] % p;
	    index[i] = j*p+index[i];
	}
    }

    /* W $@$NI=$N7W;;(J */
    windex = 0;
    p = N; q = NumberOfCell / 2; n = 1; e = 0;
    for (k = 0 ; k < 5 ; k++) {		/* $@CJ$N?t(J */
	md = N / (2*n);
	e += ((cid / q) % 2) * n / 2;
	ww[windex++] = get_w(e*md);
	p /= 2;	q /= 2;	n *= 2;
    }
    m = 1; q = N_PE / 5;
    for (k = 0 ; k < 4 ; k++) {	/* $@CJ$N?t(J */
	p /= 8; b = 0;
	md = N / (8*n);
	for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
	    e = ep[j*q]*md;
	    for (i = b ; i < b + p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		for (r = 1 ; r < 8 ; r++) {
		    ww[windex++] = get_w((e*r) % N);
		}
	    }
	    b += 8 * p;
	}
	q /= 8;	n *= 8;	m *= 8;
    }
    for (k = 0 ; k < N5_PE ; k++) {	/* $@CJ$N?t(J */
	p /= 5; b = 0;
	md = N / (5*n);
	for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
	    e = ep[j*q]*md;
	    for (i = b ; i < b + p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		for (r = 1 ; r < 5 ; r++) {
		    ww[windex++] = get_w((e*r) % N);
		}
	    }
	    b += 5 * p;
	}
	q /= 5;	n *= 5;	m *= 5;
    }
    /* $@5UJQ49(J */
    p = 1; n = N; m = N_PE;
    for (k = 0 ; k < N5_PE ; k++) {	/* $@CJ$N?t(J */
	n /= 5; m /= 5; b = 0;
	md = N / (5*p);
	for (i = 0 ; i < p ; i++) {
	    ep[i] = i * md;
	}
	for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
	    for (i = 1 ; i < p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		for (r = 1 ; r < 5 ; r++) {
		    ww[windex++] = get_w((ep[i]*r) % N);
		}
	    }
	    b += 5 * p;
	}
	p *= 5;
    }
    for (k = 0 ; k < 4 ; k++) {	/* $@CJ$N?t(J */
	n /= 8; m /= 8; b = 0;
	md = N / (8*p);
	for (i = 0 ; i < p ; i++) {
	    ep[i] = i * md;
	}
	for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
	    for (i = 1 ; i < p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		for (r = 1 ; r < 8 ; r++) {
		    ww[windex++] = get_w((ep[i]*idx[r]) % N);
		}
	    }
	    b += 8 * p;
	}
	p *= 8;
    }
    vid = (cid % (NumberOfCell / 2)) * 2;
    if (cid >= NumberOfCell / 2) {
	vid++;
    }
    q = 1;
    for (k = 0 ; k < 5 ; k++) {		/* $@CJ$N?t(J */
	md = N / (2*p);
	b = (vid % (2*q)) * N_PE / 2;
	for (i = 0 ; i < N_PE / 2 ; i++) {	/* $@3F(JPE$@Fb$NAH9g$;$N?t(J */
	    ww[windex++] = get_w(((b+i)*md) % N);
	}
	p *= 2; q *= 2;
    }
    windex = 0;

    /* $@=gJQ49(J */

    init_c(x,N_PE,&i);

    q = NumberOfCell / 2;
    if ((int)(cid / q) % 2 == 0) {
	for (i = N_PE/2 ; i < N_PE ; i++) {
	    y[i] = x[i];
	}
    } else {
	for (i = 0 ; i < N_PE/2 ; i++) {
	    y[i] = x[i];
	}
    }

    {
	/* N_PE$@8D$N%G!<%?$NH>J,$r8r49(J */
	if ((int)(cid / q) % 2 == 0) {
	    MPI_Sendrecv_replace(&y[N_PE/2],N_PE,MPI_DOUBLE,cid+q,0,cid+q,0,MPI_COMM_WORLD,&status);

	    w = ww[windex++];
	    for (i1 = 0, i2 = N_PE/2 ; i1 < N_PE / 2 ; i1++, i2++) {	/* $@3F(JPE$@Fb$NAH9g$;$N?t(J */
		z.re = w.re * y[i2].re - w.im * y[i2].im;
		z.im = w.re * y[i2].im + w.im * y[i2].re;
		X[i1].re = x[i1].re + z.re;
		X[i1].im = x[i1].im + z.im;
		X[i2].re = x[i1].re - z.re;
		X[i2].im = x[i1].im - z.im;
	    }
	} else {
	    MPI_Sendrecv_replace(y,N_PE,MPI_DOUBLE,cid-q,0,cid-q,0,MPI_COMM_WORLD,&status);

	    w = ww[windex++];
	    for (i1 = 0, i2 = N_PE/2 ; i1 < N_PE / 2 ; i1++, i2++) {	/* $@3F(JPE$@Fb$NAH9g$;$N?t(J */
		z.re = w.re * x[i2].re - w.im * x[i2].im;
		z.im = w.re * x[i2].im + w.im * x[i2].re;
		X[i1].re = y[i1].re + z.re;
		X[i1].im = y[i1].im + z.im;
		X[i2].re = y[i1].re - z.re;
		X[i2].im = y[i1].im - z.im;
	    }
	}
	q /= 2;
    }
    for (k = 1 ; k < 5 ; k++) {		/* $@CJ$N?t(J */
	if (k % 2 == 0) {
	    bufs = y; bufd = X;
	} else {
	    bufs = X; bufd = y;
	}

	/* N_PE$@8D$N%G!<%?$NH>J,$r8r49(J */
	if ((int)(cid / q) % 2 == 0) {
	    MPI_Sendrecv_replace(&bufs[N_PE/2],N_PE,MPI_DOUBLE,cid+q,0,cid+q,0,MPI_COMM_WORLD,&status);
	} else {
	    MPI_Sendrecv_replace(bufs,N_PE,MPI_DOUBLE,cid-q,0,cid-q,0,MPI_COMM_WORLD,&status);
	}

	w = ww[windex++];
	for (i1 = 0, i2 = N_PE/2 ; i1 < N_PE / 2 ; i1++, i2++) {	/* $@3F(JPE$@Fb$NAH9g$;$N?t(J */
	    z.re = w.re * bufs[i2].re - w.im * bufs[i2].im;
	    z.im = w.re * bufs[i2].im + w.im * bufs[i2].re;
	    bufd[i1].re = bufs[i1].re + z.re;
	    bufd[i1].im = bufs[i1].im + z.im;
	    bufd[i2].re = bufs[i1].re - z.re;
	    bufd[i2].im = bufs[i1].im - z.im;
	}
	q /= 2;
    }

    /* N_PE$@8D$N%G!<%?$NH>J,$r8r49(J */
    q = NumberOfCell / 2;
    if ((int)(cid / q) % 2 == 0) {
	MPI_Sendrecv_replace(&X[N_PE/2],N_PE,MPI_DOUBLE,cid+q,0,cid+q,0,MPI_COMM_WORLD,&status);
    } else {
	MPI_Sendrecv_replace(X,N_PE,MPI_DOUBLE,cid-q,0,cid-q,0,MPI_COMM_WORLD,&status);
    }

    m = 1; p = N_PE; q = N_PE / 5;
    for (k = 0 ; k < 4 ; k++) {	/* $@CJ$N?t(J */
	if (k % 2 == 1) {
	    bufs = y; bufd = X;
	} else {
	    bufs = X; bufd = y;
	}

	p /= 8; b = 0;
	for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
	    for (i = b ; i < b + p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		/* 8 $@$N(J WFT $@$N7W;;(J */

		/* w */
		for (r = 1 ; r < 8 ; r++) {
		    w = ww[windex++];
		    d = bufs[i+p*r].re * w.re - bufs[i+p*r].im * w.im;
		    bufs[i+p*r].im = bufs[i+p*r].im * w.re + bufs[i+p*r].re * w.im;
		    bufs[i+p*r].re = d;
		}

		v[1].re = bufs[i].re + bufs[i+4*p].re;	v[1].im = bufs[i].im + bufs[i+4*p].im;
		v[2].re = bufs[i+2*p].re + bufs[i+6*p].re;	v[2].im = bufs[i+2*p].im + bufs[i+6*p].im;
		v[3].re = bufs[i+p].re + bufs[i+5*p].re;	v[3].im = bufs[i+p].im + bufs[i+5*p].im;
		v[4].re = bufs[i+p].re - bufs[i+5*p].re;	v[4].im = bufs[i+p].im - bufs[i+5*p].im;
		v[5].re = bufs[i+3*p].re + bufs[i+7*p].re;	v[5].im = bufs[i+3*p].im + bufs[i+7*p].im; 
		v[6].re = bufs[i+3*p].re - bufs[i+7*p].re;	v[6].im = bufs[i+3*p].im - bufs[i+7*p].im;
		v[7].re = v[1].re + v[2].re;	v[7].im = v[1].im + v[2].im;
		v[8].re = v[3].re + v[5].re;	v[8].im = v[3].im + v[5].im;
		u[2].re = v[1].re - v[2].re;		u[2].im = v[1].im - v[2].im;
		u[3].re = bufs[i].re - bufs[i+4*p].re;	u[3].im = bufs[i].im - bufs[i+4*p].im;

		u[4].re = WFT8 * (v[4].re - v[6].re);		u[4].im = WFT8 * (v[4].im - v[6].im);
		u[5].re = v[3].im - v[5].im;		u[5].im = v[5].re - v[3].re;
		u[6].re = bufs[i+2*p].im - bufs[i+6*p].im;	u[6].im = bufs[i+6*p].re - bufs[i+2*p].re;
		u[7].re = WFT8 * (v[4].im + v[6].im);	u[7].im = - WFT8 * (v[4].re + v[6].re); 
		s[1].re = u[3].re + u[4].re;	s[1].im = u[3].im + u[4].im;
		s[2].re = u[3].re - u[4].re;	s[2].im = u[3].im - u[4].im;
		s[3].re = u[6].re + u[7].re;	s[3].im = u[6].im + u[7].im;
		s[4].re = u[6].re - u[7].re;	s[4].im = u[6].im - u[7].im;

		bufd[i].re = v[7].re + v[8].re;		bufd[i].im = v[7].im + v[8].im;
		bufd[i+p].re = v[7].re - v[8].re;	bufd[i+p].im = v[7].im - v[8].im;
		bufd[i+2*p].re = u[2].re + u[5].re;	bufd[i+2*p].im = u[2].im + u[5].im;
		bufd[i+3*p].re = u[2].re - u[5].re;	bufd[i+3*p].im = u[2].im - u[5].im;
		bufd[i+4*p].re = s[1].re + s[3].re;	bufd[i+4*p].im = s[1].im + s[3].im;
		bufd[i+5*p].re = s[2].re + s[4].re;	bufd[i+5*p].im = s[2].im + s[4].im;
		bufd[i+6*p].re = s[2].re - s[4].re;	bufd[i+6*p].im = s[2].im - s[4].im;
		bufd[i+7*p].re = s[1].re - s[3].re;	bufd[i+7*p].im = s[1].im - s[3].im;
	    }
	    b += 8 * p;
	}
	q /= 8;
	m *= 8;
    }

    for (k = 0 ; k < N5_PE ; k++) {	/* $@CJ$N?t(J */
	if (k % 2 == 1) {
	    bufs = y; bufd = X;
	} else {
	    bufs = X; bufd = y;
	}

	p /= 5; b = 0;
	for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
	    for (i = b ; i < b + p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		/* 5 $@$N(J WFT $@$r7W;;(J */

		/* w */
		for (r = 1 ; r < 5 ; r++) {
		    w = ww[windex++];
		    d = bufs[i+p*r].re * w.re - bufs[i+p*r].im * w.im;
		    bufs[i+p*r].im = bufs[i+p*r].im * w.re + bufs[i+p*r].re * w.im;
		    bufs[i+p*r].re = d;
		}

		/* r0 $@!A(J r3 $@$r7W;;$9$k(J */
		/* r0 = (W1 + W2 + W4 + W3) (x[1] + x[2] + x[3] + x[4]) */
		r0.re = - bufs[i+p].re -  bufs[i+2*p].re - bufs[i+3*p].re - bufs[i+4*p].re;
		r0.im = - bufs[i+p].im -  bufs[i+2*p].im - bufs[i+3*p].im - bufs[i+4*p].im;
		/* r1 = (W1 - W2 + W4 - W3 ) (x[1] - x[2] - x[3] + x[4]) */
		r1.re = WFT51 * (bufs[i+p].re - bufs[i+2*p].re - bufs[i+3*p].re + bufs[i+4*p].re);
		r1.im = WFT51 * (bufs[i+p].im - bufs[i+2*p].im - bufs[i+3*p].im + bufs[i+4*p].im);
		/* r2 = (W1 - W4) (x[1] - x[4]) - (W2 - W3) (x[3] - x[2]) */ 
		r2.re = - WFT52 * (bufs[i+p].im - bufs[i+4*p].im) + WFT53 * (bufs[i+3*p].im - bufs[i+2*p].im);
		r2.im = WFT52 * (bufs[i+p].re - bufs[i+4*p].re) - WFT53 * (bufs[i+3*p].re - bufs[i+2*p].re);
		/* r3 = (W1 - W4) (x[3] - x[2]) + (W2 - W3) (x[1] - x[4]) */
		r3.re = - WFT52 * (bufs[i+3*p].im - bufs[i+2*p].im) - WFT53 * (bufs[i+p].im - bufs[i+4*p].im);
		r3.im = WFT52 * (bufs[i+3*p].re - bufs[i+2*p].re) + WFT53 * (bufs[i+p].re - bufs[i+4*p].re);

		/* X0 = x0 + x1 + x2 + x3 + x4 = x0 - r0 */
		bufd[i].re = bufs[i].re - r0.re;
		bufd[i].im = bufs[i].im - r0.im;
		/* X2 = (r0 - r1 - 2*r3) / 4 + x0 */
		r3.re *= 2.0;	r3.im *= 2.0;
		bufd[i+2*p].re = (r0.re - r1.re - r3.re) / 4.0 + bufs[i].re;
		bufd[i+2*p].im = (r0.im - r1.im - r3.im) / 4.0 + bufs[i].im;
		/* X1 = (r0 + r1 - 2*r2) / 4 + x0 */
		r2.re *= 2.0;	r2.im *= 2.0;
		bufd[i+p].re = (r0.re + r1.re - r2.re) / 4.0 + bufs[i].re;
		bufd[i+p].im = (r0.im + r1.im - r2.im) / 4.0 + bufs[i].im;
		/* X3 = (r0 - r1 + 2*r3) / 4 + x0 */
		bufd[i+3*p].re = (r0.re - r1.re + r3.re) / 4.0 + bufs[i].re;
		bufd[i+3*p].im = (r0.im - r1.im + r3.im) / 4.0 + bufs[i].im;
		/* X4 = (r0 + r1 + 2*r2) / 4.0 + x0 */
		bufd[i+4*p].re = (r0.re + r1.re + r2.re) / 4.0 + bufs[i].re;
		bufd[i+4*p].im = (r0.im + r1.im + r2.im) / 4.0 + bufs[i].im;
	    }
	    b += 5 * p;
	}
	q /= 5;
	m *= 5;
    }

    /* $@@dBPCM$N:GBgCM(J */
    max1 = -1.0;
    max2 = -1.0;
    for (i = 0 ; i < N_PE ; i++) {
	t = X[i].re * X[i].re + X[i].im * X[i].im;
	if (t > max2) {
	    if (t > max1) {
	        max2 = max1; i2 = i1;
		max1 = t; i1 = i;
	    } else {
	        max2 = t; i2 = i;
	    }
        }
    }
    MPI_Allreduce(&max1,&t,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
    if (max1 == t) {
	amax[0] = X[i1];
	aindex[0] = index[i1];
	max1 = max2; i1 = i2;
    } else {
	amax[0].re = amax[0].im = 0.0;
	aindex[0] = 0;
    }
    MPI_Allreduce(&max1,&t,1,MPI_DOUBLE,MPI_MAX,MPI_COMM_WORLD);
    if (max1 == t) {
	amax[1] = X[i1];
	aindex[1] = index[i1];
    } else {
	amax[1].re = amax[1].im = 0.0;
	aindex[1] = 0;
    }
    MPI_Allreduce(amax,amax,4,MPI_DOUBLE,MPI_SUM, MPI_COMM_WORLD);
    MPI_Allreduce(aindex,aindex,2,MPI_INT,MPI_SUM, MPI_COMM_WORLD);
    half_c(amax,aindex);

	/* $@5UJQ49(J */
	p = 1; m = N_PE;
	for (k = 0 ; k < N5_PE ; k++) {	/* $@CJ$N?t(J */
	    if (k % 2 == 1) {
		bufs = y; bufd = X;
	    } else {
		bufs = X; bufd = y;
	    }

	    m /= 5; b = 0;
	    for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
		for (i = b ; i < b + p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		    /* w */
		    if (i > b) {
			for (r = 1 ; r < 5 ; r++) {
			    w = ww[windex++];
			    d = bufs[i+p*r].re * w.re + bufs[i+p*r].im * w.im;
			    bufs[i+p*r].im = bufs[i+p*r].im * w.re - bufs[i+p*r].re * w.im;
			    bufs[i+p*r].re = d;
			}
		    }

		    /* r0 $@!A(J r3 $@$r7W;;$9$k(J */
		    /* r0 = (W1 + W2 + W4 + W3) (x[1] + x[2] + x[3] + x[4]) */
		    r0.re = - bufs[i+p].re -  bufs[i+2*p].re - bufs[i+3*p].re - bufs[i+4*p].re;
		    r0.im = - bufs[i+p].im -  bufs[i+2*p].im - bufs[i+3*p].im - bufs[i+4*p].im;
		    /* r1 = (W1 - W2 + W4 - W3 ) (x[1] - x[2] - x[3] + x[4]) */
		    r1.re = WFT51 * (bufs[i+p].re - bufs[i+2*p].re - bufs[i+3*p].re + bufs[i+4*p].re);
		    r1.im = WFT51 * (bufs[i+p].im - bufs[i+2*p].im - bufs[i+3*p].im + bufs[i+4*p].im);
		    /* r2 = (W1 - W4) (x[1] - x[4]) - (W2 - W3) (x[3] - x[2]) */ 
		    r2.re = WFT52 * (bufs[i+p].im - bufs[i+4*p].im) - WFT53 * (bufs[i+3*p].im - bufs[i+2*p].im);
		    r2.im = - WFT52 * (bufs[i+p].re - bufs[i+4*p].re) + WFT53 * (bufs[i+3*p].re - bufs[i+2*p].re);
		    /* r3 = (W1 - W4) (x[3] - x[2]) + (W2 - W3) (x[1] - x[4]) */
		    r3.re = WFT52 * (bufs[i+3*p].im - bufs[i+2*p].im) + WFT53 * (bufs[i+p].im - bufs[i+4*p].im);
		    r3.im = - WFT52 * (bufs[i+3*p].re - bufs[i+2*p].re) - WFT53 * (bufs[i+p].re - bufs[i+4*p].re);

		    /* X0 = x0 + x1 + x2 + x3 + x4 = x0 - r0 */
		    bufd[i].re = bufs[i].re - r0.re;
		    bufd[i].im = bufs[i].im - r0.im;
		    /* X2 = (r0 - r1 - 2*r3) / 4 + x0 */
		    r3.re *= 2.0;    r3.im *= 2.0;
		    bufd[i+2*p].re = (r0.re - r1.re - r3.re) / 4.0 + bufs[i].re;
		    bufd[i+2*p].im = (r0.im - r1.im - r3.im) / 4.0 + bufs[i].im;
		    /* X1 = (r0 + r1 - 2*r2) / 4 + x0 */
		    r2.re *= 2.0;    r2.im *= 2.0;
		    bufd[i+p].re = (r0.re + r1.re - r2.re) / 4.0 + bufs[i].re;
		    bufd[i+p].im = (r0.im + r1.im - r2.im) / 4.0 + bufs[i].im;
		    /* X3 = (r0 - r1 + 2*r3) / 4 + x0 */
		    bufd[i+3*p].re = (r0.re - r1.re + r3.re) / 4.0 + bufs[i].re;
		    bufd[i+3*p].im = (r0.im - r1.im + r3.im) / 4.0 + bufs[i].im;
		    /* X4 = (r0 + r1 + 2*r2) / 4.0 + x0 */
		    bufd[i+4*p].re = (r0.re + r1.re + r2.re) / 4.0 + bufs[i].re;
		    bufd[i+4*p].im = (r0.im + r1.im + r2.im) / 4.0 + bufs[i].im;
		}
		b += 5 * p;
	    }
	    p *= 5;
	}

	for (k = 0 ; k < 4 ; k++) {	/* $@CJ$N?t(J */
	    if (k % 2 == 1) {
		bufs = y; bufd = X;
	    } else {
		bufs = X; bufd = y;
	    }

	    m /= 8; b = 0;
	    for (j = 0 ; j < m ; j++) {	/* $@3FCJ$NH"$N?t(J */
		for (i = b ; i < b + p ; i++) {	/* $@3FH"Fb$NAH9g$;$N?t(J */
		    /* w */
		    if (i > b) {
			for (r = 1 ; r < 8 ; r++) {
			    w = ww[windex++];
			    d = bufs[i+p*r].re * w.re + bufs[i+p*r].im * w.im;
			    bufs[i+p*r].im = bufs[i+p*r].im * w.re - bufs[i+p*r].re * w.im;
			    bufs[i+p*r].re = d;
			}
		    }

		    v[1].re = bufs[i].re + bufs[i+p].re;	v[1].im = bufs[i].im + bufs[i+p].im;
		    v[2].re = bufs[i+2*p].re + bufs[i+3*p].re;	v[2].im = bufs[i+2*p].im + bufs[i+3*p].im;
		    v[3].re = bufs[i+4*p].re + bufs[i+5*p].re;	v[3].im = bufs[i+4*p].im + bufs[i+5*p].im;
		    v[4].re = bufs[i+4*p].re - bufs[i+5*p].re;	v[4].im = bufs[i+4*p].im - bufs[i+5*p].im;
		    v[5].re = bufs[i+6*p].re + bufs[i+7*p].re;	v[5].im = bufs[i+6*p].im + bufs[i+7*p].im; 
		    v[6].re = bufs[i+6*p].re - bufs[i+7*p].re;	v[6].im = bufs[i+6*p].im - bufs[i+7*p].im;
		    v[7].re = v[1].re + v[2].re;	v[7].im = v[1].im + v[2].im;
		    v[8].re = v[3].re + v[5].re;	v[8].im = v[3].im + v[5].im;
		    u[2].re = v[1].re - v[2].re;		u[2].im = v[1].im - v[2].im;
		    u[3].re = bufs[i].re - bufs[i+p].re;	u[3].im = bufs[i].im - bufs[i+p].im;
		    u[4].re = WFT8 * (v[4].re - v[6].re);	u[4].im = WFT8 * (v[4].im - v[6].im);
		    u[5].re = v[3].im - v[5].im;		u[5].im = v[5].re - v[3].re;
		    u[6].re = bufs[i+2*p].im - bufs[i+3*p].im;	u[6].im = bufs[i+3*p].re - bufs[i+2*p].re;
		    u[7].re = WFT8 * (v[4].im + v[6].im);	u[7].im = - WFT8 * (v[4].re + v[6].re); 
		    s[1].re = u[3].re + u[4].re;	s[1].im = u[3].im + u[4].im;
		    s[2].re = u[3].re - u[4].re;	s[2].im = u[3].im - u[4].im;
		    s[3].re = u[6].re + u[7].re;	s[3].im = u[6].im + u[7].im;
		    s[4].re = u[6].re - u[7].re;	s[4].im = u[6].im - u[7].im;

		    bufd[i].re = v[7].re + v[8].re;	bufd[i].im = v[7].im + v[8].im;
		    bufd[i+4*p].re = v[7].re - v[8].re;	bufd[i+4*p].im = v[7].im - v[8].im;
		    bufd[i+6*p].re = u[2].re + u[5].re;	bufd[i+6*p].im = u[2].im + u[5].im;
		    bufd[i+2*p].re = u[2].re - u[5].re;	bufd[i+2*p].im = u[2].im - u[5].im;
		    bufd[i+7*p].re = s[1].re + s[3].re;	bufd[i+7*p].im = s[1].im + s[3].im;
		    bufd[i+3*p].re = s[2].re + s[4].re;	bufd[i+3*p].im = s[2].im + s[4].im;
		    bufd[i+5*p].re = s[2].re - s[4].re;	bufd[i+5*p].im = s[2].im - s[4].im;
		    bufd[i+p].re = s[1].re - s[3].re;	bufd[i+p].im = s[1].im - s[3].im;
		}
		b += 8 * p;
	    }
	    p *= 8;
	}

	vid = (cid % (NumberOfCell / 2)) * 2;
	if (cid >= NumberOfCell / 2) {
	    vid++;
	}

	/* N_PE$@8D$N%G!<%?$NH>J,$r8r49(J */
	q = NumberOfCell / 2;
	if ((int)(cid / q) % 2 == 0) {
	    MPI_Sendrecv_replace(&X[N_PE/2],N_PE,MPI_DOUBLE,cid+q,0,cid+q,0,MPI_COMM_WORLD,&status);
	} else {
	    MPI_Sendrecv_replace(X,N_PE,MPI_DOUBLE,cid-q,0,cid-q,0,MPI_COMM_WORLD,&status);
	}

	q = 1;
	for (k = 0 ; k < 5 ; k++) {		/* $@CJ$N?t(J */
	    if (k % 2 == 1) {
		bufs = y; bufd = X;
	    } else {
		bufs = X; bufd = y;
	    }

	    for (i1 = 0, i2 = N_PE/2 ; i1 < N_PE / 2 ; i1++, i2++) {	/* $@3F(JPE$@Fb$NAH9g$;$N?t(J */
		w = ww[windex++];
		z.re = w.re * bufs[i2].re + w.im * bufs[i2].im;
		z.im = w.re * bufs[i2].im - w.im * bufs[i2].re;
		bufd[i1].re = bufs[i1].re + z.re;
		bufd[i1].im = bufs[i1].im + z.im;
		bufd[i2].re = bufs[i1].re - z.re;
		bufd[i2].im = bufs[i1].im - z.im;
	    }

	    /* N_PE$@8D$N%G!<%?$NH>J,$r8r49(J */
	    if ((int)(cid / q) % 2 == 0) {
		MPI_Sendrecv_replace(&bufd[N_PE/2],N_PE,MPI_DOUBLE,cid+q,0,cid+q,0,MPI_COMM_WORLD,&status);
	    } else {
		MPI_Sendrecv_replace(bufd,N_PE,MPI_DOUBLE,cid-q,0,cid-q,0,MPI_COMM_WORLD,&status);
	    }

	    q *= 2;
	}

	/* X <- y / N */
	/* $@:9$N#2>hOB(J */
	t = 0.0;
	for (i = 0 ; i < N_PE ; i++) {
	    X[i].re = y[i].re / (double)N;
	    X[i].im = y[i].im / (double)N;
	    z.re = X[i].re - x[i].re;
	    z.im = X[i].im - x[i].im;
	    t += z.re * z.re + z.im * z.im;
	}
	MPI_Allreduce(&t,&sum,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
	last_c(sum);

    MPI_Finalize();

    free(W);
    free(y);
    free(x);
    free(X);
    free(ep);
    free(ew);
    free(index);
    free(ww);
}

complex get_w(int p)
{
    complex z;
    int q;

    q = p % (N/4);
    if (p < N/2) {
	if (p < N/4) {
	    z.re = W[N/4 - q];
	    z.im = - W[q];
	} else {
	    z.re = - W[q];
	    z.im = - W[N/4 - q];
	}
    } else {
	if (p < N*3/4) {
	    z.re = - W[N/4 - q];
	    z.im = W[q];
	} else {
	    z.re = W[q];
	    z.im = W[N/4 - q];
	}
    }
    return z;
}
